﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2010 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Linq;
using alglib;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using System.Collections.Generic;

namespace HeuristicLab.Problems.DataAnalysis.Operators {
  [Item("DynOpEqHistogramInitializer", "Dynamic Operator Equalization Histogram Initializer.")]
  [StorableClass]
  public class DynOpEqHistogramInitializer : SingleSuccessorOperator {
    public ILookupParameter<ItemList<IntValue>> BinCapacityParameter {
      get { return (ILookupParameter<ItemList<IntValue>>)Parameters["BinCapacity"]; }
    }
    public ILookupParameter<ItemList<ItemList<DoubleValue>>> AcceptedBinQualitiesParameter {
      get { return (ILookupParameter<ItemList<ItemList<DoubleValue>>>)Parameters["AcceptedBinQualities"]; }
    }
    public ILookupParameter<IntValue> PopulationSizeParameter {
      get { return (ILookupParameter<IntValue>)Parameters["PopulationSize"]; }
    }
    public IScopeTreeLookupParameter<SymbolicExpressionTree> SymbolicExpressionTreeParameter {
      get { return (IScopeTreeLookupParameter<SymbolicExpressionTree>)Parameters["SymbolicExpressionTree"]; }
    }
    public IScopeTreeLookupParameter<DoubleValue> QualityParameter {
      get { return (IScopeTreeLookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public ILookupParameter BinSizeParameter {
      get { return (ILookupParameter<IntValue>)Parameters["BinSize"]; }
    }
    public ILookupParameter<ItemList<IntValue>> AcceptedCountsParameter {
      get { return (ILookupParameter<ItemList<IntValue>>)Parameters["AcceptedCounts"]; }
    }
    public ILookupParameter<ItemList<IntValue>> TotalCountsParameter {
      get { return (ILookupParameter<ItemList<IntValue>>)Parameters["TotalCounts"]; }
    }

    public DynOpEqHistogramInitializer()
      : base() {
      Parameters.Add(new ScopeTreeLookupParameter<SymbolicExpressionTree>("SymbolicExpressionTree"));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("Quality"));
      Parameters.Add(new LookupParameter<IntValue>("PopulationSize"));
      Parameters.Add(new ValueLookupParameter<IntValue>("BinSize", new IntValue(5)));
      Parameters.Add(new LookupParameter<ItemList<IntValue>>("BinCapacity"));
      Parameters.Add(new LookupParameter<ItemList<ItemList<DoubleValue>>>("AcceptedBinQualities"));
      Parameters.Add(new LookupParameter<ItemList<IntValue>>("AcceptedCounts"));
      Parameters.Add(new LookupParameter<ItemList<IntValue>>("TotalCounts"));
    }

    public override IOperation Apply() {
      if (BinCapacityParameter.ActualValue == null) {
        InitDefaultCapacityHistogram();
      }
      ItemList<ItemList<DoubleValue>> acceptedBinQualities = AcceptedBinQualitiesParameter.ActualValue;
      ItemList<IntValue> acceptedCounts = AcceptedCountsParameter.ActualValue;
      ItemList<IntValue> totalCounts = TotalCountsParameter.ActualValue;

      int popSize = PopulationSizeParameter.ActualValue.Value;
      //double minQuality = (from binAccepted in acceptedBinQualities
      //                     where binAccepted.Count > 0
      //                     select binAccepted.Min(x => x.Value)).Min();
      //double maxQuality = (from binAccepted in acceptedBinQualities
      //                     where binAccepted.Count > 0
      //                     select binAccepted.Max(x => x.Value)).Max();
      //double range = maxQuality - minQuality;
      double avgQualitySum = (from binAccepted in acceptedBinQualities
                              //select binAccepted.Count)
                              where binAccepted.Count > 0
                              select (from quality in binAccepted
                                      select quality.Value).Average())
                                    .Sum();
      ItemList<IntValue> binCapacity = BinCapacityParameter.ActualValue;
      for (int i = 0; i < binCapacity.Count; i++) {
        double avgBinQuality = (from quality in acceptedBinQualities[i]
                                select quality.Value)
                               .DefaultIfEmpty(0.0)
                               .Average();

        //double avgBinQuality = acceptedBinQualities[i].Count;
        binCapacity[i].Value = (int)Math.Ceiling(popSize * (avgBinQuality / avgQualitySum));
        acceptedBinQualities[i].Clear();
        acceptedCounts[i].Value = 0;
      }
      for (int i = 0; i < totalCounts.Count; i++)
        totalCounts[i].Value = 0;
      return base.Apply();
    }

    private void InitDefaultCapacityHistogram() {
      ItemArray<SymbolicExpressionTree> trees = SymbolicExpressionTreeParameter.ActualValue;
      ItemArray<DoubleValue> quality = QualityParameter.ActualValue;
      var binCapacities = new ItemList<IntValue>();
      var acceptedQuality = new ItemList<ItemList<DoubleValue>>(20);
      var acceptedCounts = new ItemList<IntValue>();
      var totalCounts = new ItemList<IntValue>();
      BinCapacityParameter.ActualValue = binCapacities;
      AcceptedBinQualitiesParameter.ActualValue = acceptedQuality;
      TotalCountsParameter.ActualValue = totalCounts;
      AcceptedCountsParameter.ActualValue = acceptedCounts;
      for (int i = 0; i < trees.Length; i++) {
        int binIndex = GetBinIndexForSize(trees[i].Size);
        if (Exists(binIndex)) {
          AddToBin(binIndex, quality[i].Value);
        } else {
          CreateNewBin(binIndex);
          AddToBin(binIndex, quality[i].Value);
        }
      }
    }

    private void CreateNewBin(int binIndex) {
      ItemList<IntValue> binCapacities = BinCapacityParameter.ActualValue;
      ItemList<ItemList<DoubleValue>> acceptedQualities = AcceptedBinQualitiesParameter.ActualValue;
      ItemList<IntValue> acceptedCounts = AcceptedCountsParameter.ActualValue;
      ItemList<IntValue> totalCounts = TotalCountsParameter.ActualValue;
      for (int i = binCapacities.Count; i <= binIndex; i++) {
        binCapacities.Add(new IntValue(1));
        acceptedQualities.Add(new ItemList<DoubleValue>(10));
        acceptedCounts.Add(new IntValue(0));
        totalCounts.Add(new IntValue(0));
      }
    }

    private void AddToBin(int binIndex, double quality) {
      ItemList<ItemList<DoubleValue>> acceptedQualities = AcceptedBinQualitiesParameter.ActualValue;
      acceptedQualities[binIndex].Add(new DoubleValue(quality));
    }

    private bool Exists(int binIndex) {
      // if the bin has a capacity set then it exists
      ItemList<IntValue> binCapacities = BinCapacityParameter.ActualValue;
      return binIndex < binCapacities.Count;
    }

    private int GetBinIndexForSize(int size) {
      int binSize = ((IntValue)BinSizeParameter.ActualValue).Value;
      return (int)Math.Floor((size - 3.0) / binSize);
    }
  }
}
